from pydantic import BaseModel
from pydantic_settings import BaseSettings
from typing import List

class Colors(BaseModel):
    base_colors: List[str] = ['#17A2B8', "#28A745", "#0D6EFD", "#DC3545", "#FFC107"]

class Sampling(BaseModel):
    get_sampling: bool = True
    auto: bool = True
    size: float = 0.1
    auto_length: int = 100000

class Visualizations(BaseModel):
    correlation: bool = True
    missing: bool = True
    interactions: bool = True

class Settings(BaseSettings):
    colors: Colors = Colors()
    report_name: str = "Data Quality Report"
    minimal: bool = True
    file_path:str= "report.html"
    sampling: Sampling = Sampling()
    visualizations: Visualizations = Visualizations()


import warnings
import logging
import pandas as pd

from typing import List
from visions import Date, DateTime

from xurpas_data_quality.config import Settings
from xurpas_data_quality.data import get_series_descriptions, get_correlations, XTypeSet
from xurpas_data_quality.data.compare import get_compare
from xurpas_data_quality.data.descriptions import TableDescription
from xurpas_data_quality.data.alerts import get_alerts

warnings.filterwarnings('ignore', category=UserWarning)
logger = logging.getLogger(__name__)

def get_variable_type_counts(variable_types: dict)-> dict:
    value_counts = {}
    for _, value in variable_types.items():

        if value == Date or value == DateTime:
            value = 'Date'
        if value in value_counts:
            value_counts[value] += 1
        else:
            value_counts[value] = 1

    return value_counts
    

def get_overview(df: pd.DataFrame, data_types:dict=None) -> dict:
    """
    Get the overview statistics of the DataFrame.

    Args:
        df: the DataFrame object
    
    Retunr:
        dictionary object containing the table statistics
    """
    num_variables = len(df.columns)
    missing_cells = df.isnull().sum().sum()
    duplicate_rows = df.duplicated().sum()

    if data_types is not None:
        v_types = data_types

        for column in df.columns: # updates the v_types dict with the columns that are not user given data types
            if column not in v_types:
                v_types[column] = XTypeSet.infer_type(df[column])

    else:
        v_types = XTypeSet().infer_type(df)

    dataset_stats = {
        "dataset_statistics": {
            'num_variables': num_variables,
            'missing_cells': missing_cells,
            'missing_cells_perc' : (missing_cells/df.count().sum())*100,
            'duplicate_rows': duplicate_rows,
            'duplicate_rows_perc': (duplicate_rows/len(df))*100,
            'total_memory': df.memory_usage().sum(),
            'ave_memory': df.memory_usage().sum()/len(df),
            'dataset_length': len(df.index)
            },
        'variable_types': get_variable_type_counts(v_types)
    }

    return dataset_stats

def describe(df: pd.DataFrame, config:Settings,data_types: dict=None)-> TableDescription:
    """Gets Description of DataFrame which includes
    an overview, its correlations, and variable descriptions
    
    Args
        df: dataframe to be described

    Returns
        a class containing all the descriptions
    """
    logger.debug("in describe.py")
    def get_description(df, data_types, config=config):
        if config.visualizations.correlation:
            correlation = get_correlations(df)
        else:
            correlation = None
        if data_types is None:
            overview = get_overview(df)
            variables = get_series_descriptions(df)
        else:
            overview = get_overview(df, data_types)
            variables = get_series_descriptions(df, data_types)
        alerts = get_alerts(overview['dataset_statistics'], variables)
        data = TableDescription(df=df, variables=variables, correlation=correlation, alerts=alerts, **overview)

        return data

    if isinstance(df, list):
        descriptions = []
        shared_columns = [dataframe[col] for dataframe in df for col in dataframe.columns if 'volume' in col]
        shared_values = list(pd.Series(list(set(shared_columns[0]).intersection(*map(set, shared_columns[1:])))).values)
        shared_values_count = len(shared_values)
        for dataframe in df:
            data = get_description(dataframe,data_types)
            comparison_data = get_compare(dataframe, shared_values)
            data.shared_values = {'shared_values': shared_values,
                                  'shared_values_count': shared_values_count}
            data.comparison = comparison_data
            descriptions.append(data)

        return descriptions

    elif isinstance(df, pd.DataFrame):
        return get_description(df,data_types)
    

def describe_invalid(df: pd.DataFrame, errors: list, config: Settings) -> TableDescription:
    """
    Gets descriptions for the invalid dataframes/ dataframes that were ingested with error
    Args
        df: dataframe to be described

    Returns
        a class containing all the descriptions
    """
    try:
        dataset_length =   len(df.shape[0])
    except:
        dataset_length = 0
        
    dataset_stats = {
        "dataset_statistics": {
            "dataset_length": dataset_length,
            "error_counts": pd.Series(errors).value_counts().to_dict()
        }
    }


    
    return TableDescription(df=df, **dataset_stats)


def check_data(df: pd.DataFrame) -> list:
    try:
        df_empty = df[df['cbs_shipment_distribution.shipment_no'].isnull()]
        alerts = len(df_empty.index)

        if alerts <= 0:
            return

        else:
            return {"shipment_no_alerts": {"df": df_empty,
                                    "alerts": alerts}}

    except:
        return
    



import io
import base64
import matplotlib.pyplot as plt

def error_handler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(f"Error in creating plot: {e}")
            fig, ax = plt.subplots()
            ax.set(frame_on=False)
            ax.set_yticks([])
            ax.set_xticks([])
            return fig  # Replace with your actual empty plot logic

    return wrapper


def plot_to_base64_method(fig: plt.figure, minimal:bool=False):
    buf = io.BytesIO()
    
    if minimal:
        fig.savefig(buf, format="svg", bbox_inches='tight', dpi=80)
    else:    
        fig.savefig(buf, format="svg", bbox_inches='tight')
    
    plt.close(fig)

    data = base64.b64encode(buf.getvalue()).decode("utf8")
    buf.close()
    return "data:image/svg+xml;base64,{}".format(data)

def plot_to_base64(func):
    def wrapper(*args, **kwargs):
        fig = func(*args, **kwargs)
        minimal = kwargs.get('minimal', False)
        return plot_to_base64_method(fig, minimal)
    return wrapper


import pandas as pd

from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer
from xurpas_data_quality.render.render_empty import render_empty

def get_empty_report(df: pd.DataFrame, config:dict)-> HTMLBase:
    """
    Generates an empty report
    """

    body = HTMLContainer(type="sections",
                         container_items = render_empty(df, config.report_name))


    return HTMLBase(body=body, name=config.report_name)


import pandas as pd

from xurpas_data_quality.render.renderer import HTMLBase
from xurpas_data_quality.render.render_test import render_test

def get_test_report(df: pd.DataFrame, name:str)-> HTMLBase:
    """
    Generates an empty report
    """

    return render_test(df)


import pytest
import yaml
import os
import pandas as pd

import xurpas_data_quality

""" sample config data """

@pytest.fixture
def sample_df():
    data = {
        'column_1': [1, 2, 3],
        'column_2': ['A', 'B', 'C']
    }
    return pd.DataFrame(data)

@pytest.fixture
def config_file(tmp_path_factory):
    sample_config = {
        'report_name': "Config Quality Report",
        'file_path': "config_test_report.html",
        'minimal': False,
    }

    fn = tmp_path_factory.mktemp("data") / "config.yaml"
    with open(fn, 'w') as file:
        yaml.dump(sample_config,file)

@pytest.fixture
def config_file(tmpdir):
    sample_config = {
        'report_name': "Config Quality Report",
        'file_path': "config_test_report.html",
        'minimal': False,
    }
    config_path = os.path.join(tmpdir, 'empty.yaml')
    with open(config_path, 'w') as file:
        file.write("""
            colors:
            base_colors:
                - "#17A2B8"
                - "#28A745"
                - "#0D6EFD"
                - "#DC3545"
                - "#FFC107"

            report_name: "Data Quality Report"
            file_path: "report.html"
            minimal: True""")

    yield config_path

@pytest.fixture
def empty_config_file(tmpdir):
    # Create an empty YAML file
    empty_config_path = os.path.join(tmpdir, 'empty.yaml')
    with open(empty_config_path, 'w') as file:
        pass  # Empty file

    yield empty_config_path

    # Clean up: remove the temporary file
    os.remove(empty_config_path)

class TestConfigClass:
    def test_wrong_config_name(self,sample_df, capsys):
        report = xurpas_data_quality.DataReport(df=sample_df, config_file="confafadsfasf.yaml")
        captured = capsys.readouterr()
        assert "does not exist! continuing with generating report without it" in captured.out

    def test_empty_config_file(self, sample_df, capsys, empty_config_file):
        # Use the fixture to get the file path
        report = xurpas_data_quality.DataReport(df=sample_df, config_file=empty_config_file)
        captured = capsys.readouterr()
        assert "is empty! proceeding with report generation." in captured.out

    def test_config_file(self, sample_df, config_file):
        assert True


import warnings
from typing import Callable, Sequence, Set

from visions import create_type, VisionsBaseType
from visions import DateTime, Date, Generic
from visions.relations import IdentityRelation, InferenceRelation,TypeRelation
from visions.typesets import StandardSet, VisionsTypeset

import pandas as pd
from pandas.api import types as pdt

warnings.filterwarnings("ignore", category=UserWarning, module='visions.typesets.typeset')

class Numeric(VisionsBaseType):

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Generic)]
    
    @classmethod
    def contains_op(cls, series: pd.Series, state:dict)->bool:
        return pdt.is_numeric_dtype(series)
    
class Text(VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Generic)]
    
    @classmethod
    def contains_op(cls, series: pd.Series, state:dict) ->bool:
        series = series.dropna()
        return pdt.is_object_dtype(series)

def numeric_is_category(series: pd.Series, state: dict) -> bool:
    n_unique = series.nunique()
    threshold = 5
    return 1 <= n_unique <= threshold

def string_is_category(series: pd.Series, state: dict) -> bool:
    n_unique = series.nunique()
    unique_threshold = 0.5
    threshold = 50
    return (
        1 <= n_unique <= threshold and (n_unique / series.size) < unique_threshold
    )

def to_category(series: pd.Series, state: dict) -> pd.Series:
    return series.astype("category")

class Categorical(VisionsBaseType):

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [
            IdentityRelation(Generic),
            InferenceRelation(
                Numeric,
                relationship=lambda x, y: numeric_is_category(x,y),
                transformer=to_category,
            ),
            InferenceRelation(
                Text,
                relationship=lambda x, y: string_is_category(x,y),
                transformer=to_category,
            ),
        ]

    @staticmethod
    def contains_op(series: pd.Series, state: dict) -> bool:
        series = series.dropna()
        is_valid_dtype = pdt.is_categorical_dtype(series)
        if is_valid_dtype:
            return True
        return False

class XTypeSet(VisionsTypeset):
    def __init__(self):
        types = {
            Generic,
            Numeric,
            Categorical,
            Text,
            DateTime,
            }
        super().__init__(types)







import os
import yaml
import warnings
import logging
from pathlib import Path
from typing import List, Tuple

import pandas as pd

from xurpas_data_quality.report import get_report, get_empty_report, get_comparison_report, get_error_report, get_test_report
from xurpas_data_quality.data import check_dtypes, describe, load_dataframe,validate_dataframe,sample_dataframe ,check_col_names, describe_invalid, check_data, convert_to_pandas
from xurpas_data_quality.config import Settings

warnings.filterwarnings("ignore", category=UserWarning, module='visions')

logging.basicConfig(level=logging.DEBUG)
logging.getLogger('matplotlib').setLevel(logging.WARNING)
logging.getLogger('PIL').setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

class DataReport:
    def __init__(self, 
                 file:str=None, 
                 df:pd.DataFrame|List[pd.DataFrame]|Tuple[pd.DataFrame,pd.DataFrame,List] =None, 
                 report_name:str=None, 
                 file_path:str=None,
                 data_types:dict=None,
                 description: str = None,
                 minimal: bool=None,
                 sample_mode: str = 'auto',
                 sample_size: float = None,
                 auto_sample_dataset_length: int = None,
                 config_file: str = None,
                 **kwargs):
        
        logger.info("Initializing DataReport object")
        """
        Initializes the DataReport object
        Args
            file:        The path of the file you want to analyze. If empty, df parameter must exist.
                         Only supports .csv, .xlsx, .parquet, and .orc file formats
            df:          Pandas DataFrame object of data to be analyzed, If using df, file must be empty.
            report_name: Name of the report. Defaults to 'Data Quality Report'
            file_path:   Path/ directory of where the report is to be saved
            data_types:  A dictionary containing the column names and their corresponding data types.
                         If empty, then the data types will be inferred.
            minimal:     Default is True. Check if you want minimal mode as your data report.
            sample_mode: The mode of sampling. Choose between 'auto', 'manual', or 'none'. Default is 'auto'.
                         If 'Manual', you must provide a sample_size parameter. If none, no sampling is done to the data.
            sample_size: The fraction of the data to be sampled. Only used when sample_mode is 'manual'.
            auto_sample_dataset_length: The length of the dataset to be sampled. Only used when sample_mode is 'auto'. If empty, defaults to 100000.
            config_file: Optional. The path of the configuration file. If empty, default values will be used.
        """

        """Loading the config Section"""
        """The loading of config priority goes with highest priority first:
            - initialization DataReport parameters
            - config file
            - default values"""
        
        __non_config_params__ = ["self","kwargs","df","file", "config_file","sample_size", "sample_mode", "auto_sample_dataset_length","description","__non_config_params__"]
        __user_set_params__ = {key: value for key, value in locals().items() if value is not None and key not in __non_config_params__}
        logger.debug(f"__user_set_params__: {__user_set_params__}")
        if config_file != None:
            try:
                with open(config_file) as f:
                    __config_file_params__ = yaml.load(f, Loader=yaml.FullLoader)
                
                if __config_file_params__ is None:
                    raise ValueError(f"{config_file} is empty! proceeding with report generation.")

                logger.debug("Config file before being loaded in settings")
                logger.debug(__config_file_params__)
                self.config = Settings(**__config_file_params__)
                logger.debug("After loading it into Settings class")
                logger.debug(self.config)
                logger.info(f"Loaded configuration from {config_file}")

                for key, value in __user_set_params__.items():
                    setattr(self.config, key, value)

            except FileNotFoundError:
                logger.warning(f"{config_file} does not exist! continuing with generating report without it")

            except Exception as e:
                logger.error(f"Error loading config file: {e}")

            finally:
                self.__load_config__(__user_set_params__, sample_mode, auto_sample_dataset_length, sample_size)
        else:
            self.__load_init_config__(__user_set_params__=__user_set_params__,
                                       sample_mode=sample_mode, 
                                       auto_sample_dataset_length=auto_sample_dataset_length,
                                        sample_size=sample_size)

        """Checking dataframe input"""
        if description is None:
            self.dataset_description = None
        else:  
            self.dataset_description = [description]

        if isinstance(df, dict):
            df_frames = list(df.values())
            self.df = []
            for frame in df_frames:
                self.df.append(frame if isinstance(frame, pd.DataFrame) else convert_to_pandas(frame) )
                
            self.df_names = list(df.keys())
            self.render_empty = False

        elif isinstance(df, tuple):
            self.df = df[0] if isinstance(df[0], pd.DataFrame) else convert_to_pandas(df[0])
            self.df_invalid = df[1] if isinstance(df[1], pd.DataFrame) else convert_to_pandas(df[1])
            self.errors = df[2] if isinstance(df[2],list) else [df[2]]
            self.empty_df_invalid = True if not validate_dataframe(self.df_invalid, self.config) else False
            self.render_empty = True if not validate_dataframe(self.df, self.config) else False
        
        else:
            if file is not None:
                if df is not None:
                    raise KeyError("Only 'file' or 'df' should be used one at a time!")
                self.df = load_dataframe(file)
                self.render_empty = False

            else:
                if df is None:
                    raise ValueError("Please provide your data in 'file' or 'df' parameters!")
                
                self.df = df
                self.render_empty = True if not validate_dataframe(df, self.config) else False # checks if dataframe is empty

        if self.config.sampling.get_sampling:
            self.df, sample_description = sample_dataframe(self.df, self.config)
            if sample_description is not None:
                if self.dataset_description is None:
                    self.dataset_description = [sample_description]
                elif isinstance(self.dataset_description, list):
                    self.dataset_description.append(sample_description)
                else:
                    self.dataset_description = [self.dataset_description, sample_description]
                print(f"after: {self.dataset_description}")

        """checking data types"""
        if data_types is not None:
            self.data_types = check_dtypes(check_col_names(data_types, df.columns))
        else:
            self.data_types = None

        logger.debug(f"self.render_empty: {self.render_empty}")
        logger.info("DataReport object initialized successfully")

    def __load_config__(self, __user_set_params__:dict, sample_mode, auto_sample_dataset_length, sample_size):
        logger.debug(f"sample_mode: {sample_mode}")
        for key, value in __user_set_params__.items():
            self.config = self.config.model_copy(update=__user_set_params__)
        logger.debug("Config Loaded:")
        logger.debug(self.config)
        logger.debug(f"sample_mode == 'auto': {sample_mode == 'auto'}")
        sample_mode = sample_mode.lower() if sample_mode is not None else None

        """Sampling settings"""
        if sample_mode == "auto":
            logger.debug("auto sampling chosen")
            if auto_sample_dataset_length is not None:
                self.config.sampling.auto_length = auto_sample_dataset_length

        elif sample_mode == "manual":
            if sample_size is not None:
                if sample_size > 1:
                    raise ValueError("Sample size should be a fraction less than 1")
                self.config.sampling.size = sample_size
            else:
                raise ValueError("Manual sample_mode requires you to provide a sample_size (fraction between 0 and 1)")
            
        elif sample_mode is None or sample_mode == "none":
            self.config.sampling.get_sampling = False

        else:
            raise ValueError("Invalid sample_mode. Please choose between 'auto', 'manual', or 'none'")

    def __load_init_config__(self, **kwargs):
        self.config = Settings()
        self.__load_config__(**kwargs)

    def describe_dataframe(self):
        logger.info("Describing dataframe")
        self.description = describe(df=self.df, data_types=self.data_types, config=self.config)
        if self.dataset_description is not None and self.dataset_description != "":
            if self.description.description is None:
                self.description.description = []
            
            self.description.description.extend(self.dataset_description)

        logger.info("Dataframe described successfully")

    def describe_invalid_dataframe(self):
        logger.info("Describing invalid dataframe")
        self.description_invalid = describe_invalid(df=self.df_invalid, errors=self.errors, config=self.config)
        logger.info("Invalid dataframe described successfully")

    def get_data_quality_report(self, config:dict, **kwargs):
        logger.info("Generating data quality report")
        self.describe_dataframe()
        report = get_report(self.description, config)
        logger.info("Data quality report generated successfully")
        return report.render()
    
    def _render_empty_report(self, config:dict):
        logger.info("Rendering empty report")
        report = get_empty_report(self.df, config)
        logger.info("Empty report rendered successfully")
        return report.render()
    
    def _render_comparison(self, config:dict):
        logger.info("Rendering comparison report")
        self.describe_dataframe()
        report = get_comparison_report(self.description, self.df_names, config)
        logger.info("Comparison report rendered successfully")
        return report.render()
    
    def _render_error_report(self, config:dict):
        logger.info("Rendering error report")
        self.describe_invalid_dataframe()
        if self.render_empty:
            report = get_error_report(data=self.df,
                            invalid_data=self.description_invalid,
                            errors=self.errors, 
                            is_empty=self.render_empty,
                            config=config)
        
        else:
            self.describe_dataframe()
            report = get_error_report(data=self.description,
                                    invalid_data=self.description_invalid,
                                    errors=self.errors, 
                                    is_empty=self.render_empty, 
                                    config=config)
        logger.info("Error report rendered successfully")
        return report.render()
    
    def __render_test_report(self):
        logger.info("Rendering test report")
        self.describe_dataframe()
        report = get_test_report(df=self.description, name=None)
        rendered_report = report.render(self.config)
        logger.info("Test report rendered successfully")
        return rendered_report

    def to_file_test(self):
        logger.info(f"Saving test report to {self.config.file_path}")
        output = Path(self.config.file_path)
        output.write_text(self.__render_test_report(), encoding='utf-8')
        logger.info("Test report saved successfully")

    def to_file(self):
        logger.info(f"Saving report to {self.config.file_path}")
        output = Path(self.config.file_path)
        if hasattr(self, 'errors'):
            logger.info(f"saving error report as {self.config.file_path}")
            output.write_text(self._render_error_report(config=self.config), encoding='utf-8')
            logger.info(f"saved!")

        elif self.render_empty:
            logger.info(f"saving empty report as {self.config.file_path}")
            output.write_text(self._render_empty_report(config=self.config), encoding='utf-8')
            logger.info("saved!")

        elif isinstance(self.df, list):
            logger.info(f"saving comparison report!")
            output.write_text(self._render_comparison(config=self.config), encoding='utf-8')
            logger.info('saved!')
        
        else:
            logger.info(f"saving as {self.config.file_path}")
            if self.config.minimal:
                logger.info("saving minimal version of report!")
                from minify_html import minify
                minified_report = minify(self.get_data_quality_report(config=self.config))
                output.write_text(minified_report, encoding='utf-8')
            else:
                output.write_text(self.get_data_quality_report(config=self.config), encoding='utf-8')
            
            logger.info(f"saved!")


import pandas as pd

from visions import DateTime, Date

from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.render.render_types import render_numerical, render_categorical, render_date, render_generic, render_string
from xurpas_data_quality.data.descriptions import TableDescription
from xurpas_data_quality.data.typeset import Numeric, Categorical, Text
from xurpas_data_quality.visuals import plot_to_base64,create_heatmap, create_interaction_plot, create_missing_bar_plot

def render_variable(data: dict, column: str, config:dict):
    if data['type'] == Numeric:
        return render_numerical(data, column, config)
    
    elif data['type'] == Categorical:
        return render_categorical(data, column, config)
    
    elif data['type'] == Date or data['type'] == DateTime:
        return render_date(data, column, config)
    
    elif data['type'] == Text:
        return render_string(data, column, config)
    
    else:
        return render_generic(data, column, config)
    
def render_variables_section(data: TableDescription, config:dict, **kwargs) -> list:
    vars = []
    for key, value in data.variables.items():
        variable = render_variable(value, key, config)
        vars.append(variable)
    return vars

def render_missing_section(data: pd.DataFrame, config:dict):
    if config.visualizations.missing:
        return HTMLContainer(
            type="box",
            name="Missing",
            container_items=[
                HTMLPlot(plot=create_missing_bar_plot(data.df, minimal=config.minimal),
                        type="large",
                        id="missingplot",
                        name="Missing Bar Plot")]
        )

def render_correlation_section(data: pd.DataFrame, config:dict):
    if config.visualizations.correlation:
        return HTMLContainer(
            type="box",
            name="Correlation",
            container_items=[
                HTMLContainer(
                    type="tabs",
                    container_items=[
                        HTMLPlot(plot=create_heatmap(data, minimal=config.minimal),
                                type="large",
                                id="corr",
                                name="Heatmap"),
                        HTMLTable(
                            id='sample',
                            name="Table",
                            data=data.to_html(classes="table table-sm", border=0, justify='left', index=False))
                    ]
                )
            ]
        )

def render_interactions_section(data:pd.DataFrame, config:dict)->HTMLBase:
    if config.visualizations.interactions:
        df = data.select_dtypes(exclude=['object'])
        outer_tabs = []
        for column in df.columns:
            inner_tabs = []
            for inner_col in df.columns:
                inner_tabs.append(
                    HTMLContainer(
                        type="default",
                        name= inner_col,
                        id = f"{column}-{inner_col}-interaction-inner",
                        container_items = [HTMLPlot(
                            plot= create_interaction_plot(df[inner_col],df[column], minimal=config.minimal),
                            type = "large",
                            name = f"{column}-{inner_col} Interaction Plot",
                            id = f"{column}-{inner_col}_interaction_plot"
                        )]
                    )
                )
            outer_tabs.append(
                HTMLContainer(
                    type="tabs",
                    name = column,
                    id = f"{column}-interaction-outer",
                    container_items = inner_tabs
                )
            )
        return outer_tabs

def render_dropdown_section(items: HTMLBase, names:list, config:dict, id="variables-dropdown" )-> list:
    return [
        HTMLDropdown(
        dropdown_items= names,
        dropdown_content= HTMLContainer(
            type="default",
            container_items= items),
        id=id
        )
    ]


import os
import warnings

import pandas as pd
from pyspark.sql.types import StringType
from pyspark.sql.functions import to_timestamp
from pyspark.sql.functions import year, col

from xurpas_data_quality.config import Settings

def load_csv(file_path, **kwargs):
    return pd.read_csv(file_path)

def load_excel(file_path, **kwargs):
    return pd.read_excel(file_path, **kwargs)

def load_parquet(file_path, **kwargs):
    return pd.read_parquet(file_path, **kwargs)

def load_orc(file_path, **kwargs):
    return pd.read_orc(file_path, **kwargs)

def sample_dataframe(df: pd.DataFrame, config: Settings):
    """
    Returns a sample of the dataframe.

    Args
        df: dataframe to sample
        sample_size: percentage of the sample

    Returns
        A sample of the dataframe
    """

    if config.sampling.auto:
        if df.shape[0] > config.sampling.auto_length:
            sampled_df = df.sample(frac=config.sampling.size)
            description = f"Report created with a sample of {config.sampling.size*100}% of the data"
        else:
            sampled_df = df
            description = None
    else:
        sampled_df = df.sample(frac=config.sampling.size)
        description = f"Report created with a sample of {config.sampling.size*100}% of the data"

    return sampled_df, description

def validate_dataframe(df: pd.DataFrame, config: Settings) -> bool:
    """
    Checks if dataframe is empty.

    Args
        df: dataframe to validate

    Returns
        A boolean. False if the dataframe is empty, True if otherwise
    """
    if df.empty:
        warnings.warn("DataFrame is empty! Generating Empty Report")
        return False
    else:

        return True

def load_dataframe(file_path: str) -> pd.DataFrame:
    load_methods = {
    '.csv': load_csv,
    '.xlsx': load_excel,
    '.parquet': load_parquet,
    '.orc': load_orc
    }
    # Get the file extension
    _, file_extension = os.path.splitext(file_path)
    
    # Get the loading method from the dictionary
    load_method = load_methods.get(file_extension)
    
    # If the loading method exists, call it, else raise an error
    if load_method:
        return load_method(file_path)
    else:
        raise ValueError(f'Unsupported file format: {file_extension}')

def convert_to_pandas(spark_df):
    """
    Converts to pandas dataframe the spark_df
    """
    try:
        df = spark_df.toPandas()


    except pd.errors.OutOfBoundsDatetime as e:
        print(f"ERROR: {e}")
        for column, dtype in spark_df.dtypes:
            print(f"{column}: {dtype}")
            if dtype =='timestamp':
                spark_df = spark_df.withColumn(column, spark_df[column].cast(StringType()))
                spark_df = spark_df.withColumn(column, to_timestamp(spark_df[column], 'yyyy-MM-dd HH:mm:ss'))

        try:
            df = spark_df.toPandas()

        except:
            for column, col_dtype in spark_df.dtypes:
                if str(col_dtype) == "timestamp":
                    filtered_df = spark_df.filter(year(col(column)) >= 2262)

            df = filtered_df.toPandas()

    return df

    



import logging
logger = logging.getLogger(__name__)
from typing import Any, List

from xurpas_data_quality.data.descriptions import TableDescription
from xurpas_data_quality.render.renderer import BaseRenderer, HTMLContainer, HTMLTable, HTMLPlot, HTMLText
from xurpas_data_quality.render.render_sections import (render_variables_section, 
                                                    render_missing_section,
                                                    render_correlation_section, 
                                                    render_interactions_section, 
                                                    render_dropdown_section)

def  create_overview_section(data: TableDescription, dataset_statistics: dict) -> HTMLContainer:
    

    overview = HTMLContainer(
            type="default",
            name="Overview",
            id="overview-tab",
            container_items=[
                HTMLContainer(
                    type="column",
                    container_items=HTMLTable(
                        data=dataset_statistics,
                        name="Dataset Statistics"
                    )),
                HTMLContainer(
                    type="column",
                    container_items=HTMLTable(
                        data=data.variable_types,
                        name="Variable Types"
                    )
                )
            ]
        )
    
    overview_contents = [overview]

    if data.description is not None:
        description = HTMLContainer(
            type="default",
            name="Description",
            id="description-tab",
            container_items=[
                HTMLText(
                    text = data.description
                )
            ]
        )

        overview_contents.append(description)

    '''if data.alerts is not None and "shipment_no_alerts" in data.alerts:
        alerts = HTMLContainer(
            type="default",
            name="Alerts",
            id="alerts-tab",
            flex=True,
            container_items=[
                HTMLTable(
                    id="alerts",
                    data=[f"There are {data.alerts['shipment_no_alerts']['alerts']} rows with no Shipment Numbers"]
                ),
                HTMLTable(
                    id="alert_sample",
                    data=data.alerts["shipment_no_alerts"]["df"].to_html(classes="table table-sm", border=0, justify='left')
                )
            ]
        )
        overview_contents.append(alerts)'''

    return HTMLContainer(
        type="box",
        name="Overview",
        id="overview_section",
        container_items=[
            HTMLContainer(
                type="tabs",
                container_items=overview_contents
            )
        ]
    )

def render_report(data: TableDescription, config: dict) -> List[BaseRenderer]:
    """
    Creates the content of a normal report
    """
    content = []

    dataset_statistics = {
        'Length of Dataset': data.dataset_statistics['dataset_length'],
        'Number of Variables': data.dataset_statistics['num_variables'],
        'Missing Cells': data.dataset_statistics['missing_cells'],
        'Missing Cells (%)': "{:0.2f}%".format(data.dataset_statistics['missing_cells_perc']),
        'Duplicate Rows': data.dataset_statistics['duplicate_rows'],
        'Duplicate Rows (%)': "{:0.2f}%".format(data.dataset_statistics['duplicate_rows_perc'])
    }

    overview_section = create_overview_section(data, dataset_statistics)
    content.extend([overview_section])

    samples_section = HTMLContainer(
        type="box",
        name="Sample",
        container_items=[
            HTMLTable(
                id="sample",
                data=data.df.head(10).to_html(classes="table table-sm", border=0, index=False, justify='left')
            )
        ]
    )
    content.extend([samples_section])

    variables_section = HTMLContainer(
        type="box",
        name="Variables",
        container_items=render_dropdown_section(items=render_variables_section(data, config=config), names=list(data.df), config=config)
    )
    content.extend([variables_section])

    if config.visualizations.missing:
        missing_section = render_missing_section(data, config)
        content.extend([missing_section])

    if config.visualizations.correlation:
        correlation_section = render_correlation_section(data.correlation, config)
        content.extend([correlation_section])

    if not config.minimal:
        if config.visualizations.interactions:
            interactions_section = HTMLContainer(
                type="box",
                name="Interactions",
                container_items=[
                    HTMLContainer(
                        type="tabs",
                        container_items=render_interactions_section(data.df, config)
                    )
                ]
            )
            content.extend([interactions_section])
    
    return content


from typing import List

import pandas as pd

def get_compare(df: pd.DataFrame, shared_values:list):
    # for every df, get, number of unique values, number of values per table, what values exist in other tables
    def handle_error(func, *args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(f"Error occurred: {e}")
            return "#N/A"
        
    if df.empty:
        unshared_df = pd.DataFrame()
        value_counts = "#N/A"
        unique_counts = "#N/A"
        unique_counts_perc = "#N/A"
        unshared_values = "#N/A"
    
    else:

        try:
            volume_columns = df.columns[df.columns.str.contains('volume', case=False)]
        except:
            volume_columns = "#N/A"
        
        try:
            value_counts = len(df[volume_columns].index)
        except:
            value_counts = "#N/A"

        try:
            unique_counts = df[volume_columns].nunique(dropna=True).iloc[0]
        except:
            unique_counts = "#N/A"
        
        try:
            unique_counts_perc = (unique_counts/value_counts)*100
        except:
            unique_counts_perc = "#N/A"
        
        try:
            unshared_values_series = unshared_df[volume_columns].dropna()
        except:
            unshared_values_series = "#N/A"

        try:
            unshared_values = unshared_values_series.squeeze().to_list()  if isinstance(unshared_values_series.squeeze(), pd.Series) else unshared_values_series.squeeze()
        except:
            unshared_values = "#N/A"
        
        try:
            unshared_values_rows = unshared_df.dropna(subset=['volume']).index
        except:
            unshared_values_rows = "#N/A"

        try:
            unshared_df = df.loc[unshared_values_rows]
        except:
            unshared_df = pd.DataFrame()

    
    return {
        'df': df,
        'unshared_df':unshared_df,
        'value_count':value_counts,
        'distinct_count':unique_counts,
        'distinct_perc': unique_counts_perc,
        'unshared_values': unshared_values
    }


import pandas as pd
import numpy as np
import string

from typing import  Dict, Tuple

def _get_variable_frequencies(series: pd.Series, num_rows: int = 10) -> pd.DataFrame:
    series_value_counts = series.value_counts()

    # Create a DataFrame from the value counts
    df_value_counts = pd.DataFrame({'Count': series_value_counts.values}, index=series_value_counts.index)

    # Calculate the frequency and add it to the DataFrame
    df_value_counts['Frequency (%)'] = (df_value_counts['Count'] / series_value_counts.sum()) * 100

    # Add a row for "Other values"
    if len(series_value_counts) > num_rows:
        other_values_count = sum(series_value_counts.values[num_rows:])
        other_values_freq = (other_values_count / series_value_counts.sum()) * 100
        other_values_df = pd.DataFrame({'Count': [other_values_count], 'Frequency (%)': [other_values_freq]}, index=['Other values (' + str(len(series_value_counts) - num_rows) + ')'])
        df_variable_frequencies = pd.concat([df_value_counts.head(num_rows), other_values_df])
    else:
        df_variable_frequencies = df_value_counts.head(num_rows)

    # Set the name of the index
    df_variable_frequencies.index.name='Value'
    df_value_counts.index.name='Value'

    # Round the 'Frequency (%)' column to two decimal places and add a percentage sign
    df_variable_frequencies['Frequency (%)'] = df_variable_frequencies['Frequency (%)'].apply(lambda x: f'{x:.2f}%')

    return df_variable_frequencies

def _get_word_count(series: pd.Series) -> pd.Series:
    word_lists = series.str.lower().str.split()
    words = word_lists.explode().str.strip(string.punctuation+string.whitespace)
    word_counts = pd.Series(words.index, index=words)
    word_counts = word_counts[word_counts.index.notnull()]
    word_counts = word_counts.groupby(level=0, sort=False).sum()
    word_counts = word_counts.sort_values(ascending=False)

    return word_counts

def _get_samples(series: pd.Series):
    first_five = series.head(5)

    series_samples = {f"row_{value}": value for value in first_five }

    return series_samples

def describe_string(series: pd.Series, summary: Dict) ->  Tuple[pd.Series, Dict]:
    """
    Describes a string series.

    Args
        series: a pandas Series with a Numeric dtype to be described
        summary: dict containing the series descriptions so far

    Returns
        The series and the updated summary

    """
        # make sure that the series has no numbers

    def _get_percentage(divisor: int, dividend=summary['value_counts']):
        try:
            return (divisor / dividend) * 100
        except:
            return 0
    
    series = series.astype('string')
    
    max_length = series.str.len().max()
    min_length = series.str.len().min()
    median_length = series.str.len().median()
    mean_length = series.str.len().mean()
    unique = sum(summary['value_counts']==1)
    unique_perc = _get_percentage(unique)
    sample = _get_samples(series)
    word_counts = _get_word_count(series)

    summary.update({
        "max_length": max_length,
        "min_length": min_length,
        "median_length": median_length,
        "mean_length": mean_length,
        "unique": unique,
        "unique_perc": unique_perc,
        "histogram_counts": series.value_counts().to_dict(),
        "samples": sample,
        "word_counts": word_counts,
        "variable_frequencies": _get_variable_frequencies(series)
    })

    return summary
